/****************************************************************************
 * QLogic ISP2x00 Fibre Channel driver
 * Copyright (C) 2000 Qlogic Corporation 
 * (www.qlogic.com)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 ****************************************************************************/

/*
 * vendor specific op codes.
*/

#define	UCSCSI_DCMD			0x20 /* DAC960 direct command */
#define DAC_CDB_LEN			12
#define DAC_SENSE_LEN			64

#define DACMD_WRITE_CONF_ONDISK		0x4B
#define	DACMD_WRITE_CONFIG		0x06
#define	DACMD_WRITE_CONF2		0x3C
#define	DACMD_WRITE_CONFLABEL		0x49 /* Write configuration label */
#define	DACMD_WRITE_CONFIG_V3x		0x4F
#define	DACMD_ADD_CONFIG_V2x		0x18
#define	DACMD_ADD_CONFIG_V3x		0x4C
#define	DACMD_STORE_IMAGE		0x21
#define	DACMD_ADD_CAPACITY		0x2A /* add physical drives to existing array */
#define	DACMD_WRITE_IOPORT		0x3A /* write port B */
#define	DACMD_S2S_WRITEFULLCONF		0x60 /* write full configuration */
#define	DACMD_S2S_ADDFULLCONF		0x62 /* add   full configuration */
#define	DACMD_S2S_WRITELUNMAP_OLD	0x58 /* write LUN map information */
#define DACMD_S2S_WRITELUNMAP		0xD2 /* Write LUN MAP Information */
#define	DACMD_S2S_WRITE_IOPORT		0x66 /* write expanded IO port */
#define	DACMD_WRITE_V3x			0x34 /* write data from plain memory */
#define	DACMD_S2S_WRITESIG		0x4D /* write signature information */

    /*
    * qla2100_set_scsi_direction
    *      This routine will set the proper direction for vendor specific
    *      commands. 
    *
    *      Note: Vendors should modify this routine to set the proper 
    *      direction of the transfer if they used vendor specific commands.
    *
    * Input:
    *      ha = adapter block pointer.
    *      sp = SCSI Request Block structure pointer.
    *
    * Returns:
    *      0 = success, was able to issue command.
    */
    STATIC void
    qla2100_set_vend_direction(scsi_qla_host_t *ha, 
 			Scsi_Cmnd       *cmd, cmd_entry_t     *pkt) {
            
			if( cmd->data_cmnd[0] == UCSCSI_DCMD )
			{
			    pkt->control_flags = 0;
			    switch( cmd->data_cmnd[2] )
			    {
				case DACMD_WRITE_CONF_ONDISK:
				case DACMD_WRITE_CONFIG:
				case DACMD_WRITE_CONF2:
				case DACMD_WRITE_CONFLABEL:
				case DACMD_WRITE_CONFIG_V3x:
				case DACMD_ADD_CONFIG_V2x:
				case DACMD_ADD_CONFIG_V3x:
				case DACMD_STORE_IMAGE:
				case DACMD_ADD_CAPACITY:
				case DACMD_WRITE_IOPORT:
				case DACMD_S2S_WRITEFULLCONF:
				case DACMD_S2S_ADDFULLCONF:
				case DACMD_S2S_WRITELUNMAP_OLD:
				case DACMD_S2S_WRITELUNMAP:
				case DACMD_S2S_WRITE_IOPORT:
				case DACMD_WRITE_V3x:
				case DACMD_S2S_WRITESIG:
					pkt->control_flags |= BIT_6;
					break;
				default:
					pkt->control_flags |= BIT_5;
			    }
			}
			else
			{
                       		pkt->control_flags |= BIT_5;
			}

    }

